\
\ Code for the DCC generator.
\
\ $ProjectVersion: 0.98 $
\

include picisr.fs

\ ----------------------------------------------------------------------
\ Pinout used in this program
\ ----------------------------------------------------------------------

0 pin-b dcc-output
1 pin-b led-empty
2 pin-b led-error

\ ----------------------------------------------------------------------
\ Store and fetch from memory banks
\ ----------------------------------------------------------------------

code bank-get-prepare ( addr -- x )
    fsr ,w movf               \ Get stack pointer
    tmp1 movwf                \ and save it
    loadw                     \ Get address,
    fsr movwf                 \ prepare to read value
    return
end-code

code bank-get-terminate ( x -- n )
    tmp2 movwf                \ Save read value
    tmp1 ,w movf              \ Get old stack pointer
    fsr movwf                 \ and restore it
    tmp2 ,w movf              \ Get saved value
    indf movwf                \ Write the read value
    return
end-code

code bank1-get ( addr -- n )
    bank-get-prepare call
    status reg-rp0 bsf        \ Choose bank 1
    indf ,w movf              \ read value
    status reg-rp0 bcf        \ Restore bank 0
    bank-get-terminate goto
end-code

code bank2-get ( addr -- n )
    bank-get-prepare call
    status reg-rp1 bsf        \ Choose bank 2
    indf ,w movf              \ read value
    status reg-rp1 bcf        \ Restore bank 0
    bank-get-terminate goto
end-code

code bank-set-prepare ( v addr -- )
    popw                      \ Get address
    tmp1 movwf                \ and save it
    popw                      \ Get value
    tmp2 movwf                \ and save it
    fsr ,w movf               \ Get stack pointer
    tmp3 movwf                \ and save it
    tmp1 ,w movf              \ Get address
    fsr movwf                 \ and prepare to write value
    tmp2 ,w movf              \ Get value
    return
end-code

code bank1-set ( v addr -- )
    bank-set-prepare call
    status reg-rp0 bsf        \ Choose bank 1
    indf movwf                \ Write value
    status reg-rp0 bcf        \ Restore bank 0
    tmp3 ,w movf              \ Get old stack pointer
    fsr movwf                 \ and restore it
    return
end-code

code bank2-set ( v addr -- )
    bank-set-prepare call
    status reg-rp1 bsf        \ Choose bank 2
    indf movwf                \ Write value
    status reg-rp1 bcf        \ Restore bank 0
    tmp3 ,w movf              \ Get old stack pointer
    fsr movwf                 \ and restore it
    return
end-code

variable current-bank1        \ True if working in bank 1, false if in bank 2

: bank1 ( -- ) 1 current-bank1 ! ;
: bank2 ( -- ) 0 current-bank1 ! ;

: bank-get ( addr -- n ) current-bank1 @ if bank1-get exit then bank2-get ;
: bank-set ( v addr -- ) current-bank1 @ if bank1-set exit then bank2-set ;

variable src
variable dst

\ Increment src and dst

: inc-vars ( -- ) 1 src +! 1 dst +! ;

\ Copy one byte from src to dst in the selected bank

: to-bank ( -- ) src @ @ dst @ bank-set ;

\ Copy one byte from src in the selected bank to dst

: from-bank ( -- ) src @ bank-get dst @ ! ;

\ ----------------------------------------------------------------------
\ Serial port handling
\ ----------------------------------------------------------------------

: handle-serial ( -- ) ;    \ XXXXX Not yet implemented

\ ----------------------------------------------------------------------
\ Interrupt driven DCC pulse generator
\ ----------------------------------------------------------------------

\ set-timer will wait for 3.2*(256-n)-0.2 s for a 20MhZ oscillator and a
\ 16 prescaler. For ones, use n=238 (57.4s => 1.03% error). For zero, use
\ values between n=0 (819.0s) and n=224 (102.2s). If necessary, the range
\ will be adjusted by testing the timer a different

:: set-timer ( n -- ) tmr0 +! ;

\ pulse and half-pulse work the reverse way around: they set the timer
\ to the proper value, then call the idle-loop to do other tasks, then
\ wait for the timer to expire and immediately change the DCC output
\ state.  This way, it is permitted to take some time (up to 40s,
\ i.e. around 200 instructions) to handle secondary tasks. More
\ complex tasks can be handled if the zero flag is true (in which case
\ we have up to 80s, i.e. around 400 instructions).

: half-pulse ( n -- )
    clrwdt                     \ Clear watchdog counter at least every 18ms
    set-timer ;                \ Set timer to pulse value

\ isr-exit is a word doing isr-restore-return

code isr-exit ( -- )
    isr-restore-return
end-code

\ low-delay contains the duration of the next low pulse when the high pulse
\ is in effect, and 0 otherwise
\ Warning: those words (pulse, bit-zero and bit-one) terminate with a call
\ to isr-exit, thus returning from interrupt

variable low-delay

: pulse ( low high -- ) half-pulse low-delay ! isr-exit ;

: bit-zero ( -- ) $64 dup pulse ;  \ Could implement zero stretching
: bit-one ( -- )  $ee dup pulse ;

\ Data buffer to be transmitted. As soon as the length is zero, it is
\ possible to write another two data bytes, then the length. Since the
\ DCC pulse generator moves the data in its internal buffer, a task
\ has virtually the length of a byte plus a stop or a start bit to
\ prepare its data to be sent.

variable transmit1
variable transmit2
variable transmit-length

\ Data being transmitted. We implement a 16 bits shift register, left-aligned,
\ so that preamble, start and stop bits can be easily implemented.

variable data1
variable data2
variable data-length

\ isr-tmr0 is called when there is a timer overflow interrupt
\ This word is very long to avoid return stack depth exhaustion, but very
\ thoroughly commented. It terminates with a call to isr-exit to properly
\ exit from interrupt

: isr-tmr0 ( -- )
    t0if intcon bit-clr                 \ Acknowledge interrupt
    low-delay @ if
	dcc-output low                  \ Drive DCC output low
	low-delay @                     \ Get low pulse delay
	0 low-delay !                   \ Note low pulse in progress
	half-pulse                      \ Reset timer and clear watchdog
	isr-exit exit                   \ Properly terminate isr
    then
    dcc-output high                     \ Drive DCC output high
    begin                               \ Find next bit to transmit
	data-length @ if 	        \ If there is another bit, transmit it
	    -1 data-length +!           \ Decrement data length
	    ]asm
	    data2 ,f rlf                \ Shift data bytes one bit
	    data1 ,f rlf                \ left through carry
	    status c btfsc              \ Skip if carry clear (0 to transmit)
	    bit-one goto                \ Transmit one
	    bit-zero goto               \ Transmit zero
	    asm[
	then
	transmit-length @ if            \ If there is data to transmit
	    transmit1 @ data1 !         \ Copy first byte
	    transmit2 @ data2 !         \ Copy second byte
	    transmit-length @ data-length !  \ Copy data length
	    0 transmit-length !         \ Reset transmit length
	    led-empty low               \ Clear "nothing-to-transmit" led
	else                            \ No data is ready, transmit preamble
	    $ff data1 !                 \ Preamble is only ones
	    $c0 data2 !                 \ (10 bits)
	    $a data-length !
	    led-empty high              \ Set nothing-to-transmit led
	then
    again                               \ Retry with new data installed
    isr-exit exit
;

\ ----------------------------------------------------------------------
\ DCC data manipulation
\ ----------------------------------------------------------------------

variable checksum

\ Wait for the transmission slot to be available

: wait-for-transmission ( -- )
    begin transmit-length @ while repeat ;

\ Transmit a preamble (10 bits to 1) and a start bit (one bit to 0)

: preamble ( -- )
    wait-for-transmission
    0 checksum !
    $ff transmit1 ! $c0 transmit2 ! $b transmit-length ! ;

\ send-byte sends a byte followed by a start bit. send-final-byte ends
\ the packet with a stop bit after the byte. send-checksum sends the checksum
\ with a stop bit.

: finish-byte ( b -- )
    dup checksum xor! transmit1 ! 9 transmit-length ! ;

: send-byte ( b -- )
    wait-for-transmission $00 transmit2 ! finish-byte ;

: send-final-byte ( b -- )
    wait-for-transmission $80 transmit2 ! finish-byte ;

: send-checksum ( -- ) checksum @ send-final-byte ;

\ Short loco speed command

: send-loco-speed ( s n -- )
    preamble send-byte send-byte send-checksum ;

\ ----------------------------------------------------------------------
\ Initialisation
\ ----------------------------------------------------------------------

: init-portb ( -- )
    $00 portb !        \ Clear output latches
    $f8 trisb !        \ Ports 0, 1 and 2 as output
    \ Interrupts should be configured here
;

: init-timer ( -- )
    $03 option_reg !             \ CLKOUT, prescaler to timer 0, prescaler=16
    t0ie intcon bit-set          \ Enable timer 0 overflow interrupt
    $f0 tmr0 !                   \ Do not wait too long before first interrupt
    ;

: init-bank1 ( -- ) 0 a0 bank1-set ;

: init-serial ( -- )
    $04 txsta !                  \ asynchronous no-tx high-speed
    $90 rcsta !                  \ serial-enable 8bits-rx continuous-receive
    $81 spbrg !                  \ 9600 bauds with BGRH high at 20MhZ
;

: init-dcc ( -- )
    0 data-length !              \ no data being transmitted
    0 transmit-length !          \ no data queued for transmission
;
    
: init ( -- )
    init-dcc init-portb init-bank1 init-serial init-timer enable-interrupts
;

\ ----------------------------------------------------------------------
\ Global interrupt handler
\ ----------------------------------------------------------------------

isr
: isr-handler ( -- )
    isr-save
    t0if intcon bit-set? if isr-tmr0 exit then
    led-error high
    isr-exit
;

\ ----------------------------------------------------------------------
\ Main program and main loop
\ ----------------------------------------------------------------------

: mainloop ( -- )
    begin $68 $03 send-loco-speed again ;

main
: main ( -- ) init mainloop ;

\ ----------------------------------------------------------------------
\ Configuration word
\ ----------------------------------------------------------------------

fosc-hs set-fosc    \ High-speed oscillator
