\
\ PicForth $ProjectVersion: 0.98 $
\
\ Forth compiler for Microchip PIC 16F873 / 16F876
\
\ Copyright (c) 2002 Samuel Tardieu <sam@rfc1149.net>
\
\ This compiler is released under the GNU General Public License version 2,
\ see the COPYING file in the same directory, as well as the README.
\

hex
\ ----------------------------------------------------------------------
\ PicForth name spaces
\ ----------------------------------------------------------------------

\ We do use several vocabularies in PicForth:
\   + picassembler: contains the postfix/prefix PIC assembler
\   + picforth:     contains the code compiled for the PIC
\   + metacompiler: contains code needed for meta compilation, such as
\                   immediate words

vocabulary picassembler
vocabulary picforth
vocabulary metacompiler

\ We do use several modes here:
\   + picasm
\       In this mode, definitions go the picassembler vocabulary. forth words
\       are visible, as well as picassembler ones and then metacompiler ones.
\   + meta
\       In this mode, definitions go to the metacompiler vocabulary. forth
\       words are visible, as well as metacompiler ones. picassembler words
\       are next, so that immediate words can be defined using the assembler,
\       then the picforth word list to ease reference to already defined
\       words. Care must be taken in this mode as newly defined words may be
\       hidden by existing forth words. In short, meta should only be used
\       to define words that can be seen when compiling for the target.
\   + target
\       In this mode, we do compile for the target. Target words are seen
\       first (XXXXX maybe invert with metacompiler), then metacompiler
\       ones. Note that forth words are not visible to prevent mistakes
\       (such as using a word that has not be defined in picforth but iss
\       present in the host's vocabulary).
\   + host
\       Regular mode: forth dictionary is first and inserted into, then
\       the picforth and picassembler dictionary.

: picasm only metacompiler also picassembler definitions also forth ;
: meta only picforth also picassembler also metacompiler definitions
    also forth ;
: target only metacompiler also picforth definitions ;
: host only picassembler also picforth also forth definitions ;

\ Import a word from the forth dictionary into the metacompiler wordlist
\ (has to be executed in meta mode)

: import: ( "name" -- ) >in @ >r : r> >in ! name evaluate postpone ; ;

\ ----------------------------------------------------------------------
\ Memory allocation
\ ----------------------------------------------------------------------

\ We are not being strict here. At this time, we assume that all the
\ code fits in 14kbits (10 bits addressing) and that data fits in one
\ bank (except special registers).

host

2000 cells allocate throw constant tcode
1 value tcshere
: tcs! cells tcode + ! ;
: tcs@ cells tcode + @ ;

\ This space contains copies of literals with all significant bits

2000 cells allocate throw constant tlit
: tcslit! cells tlit + ! ;
: tcslit@ cells tlit + @ ;

\ Data area (RAM)

2000 cells allocate throw constant tdata
20 value there
: t! tdata + ! ;
: check-data-area there 80 >= abort" out of RAM space" ;
: dataorg to there check-data-area ;

\ EEPROM

2000 cells allocate throw constant teeprom
0 value teehere
: tee! teeprom + c! ;
: tee@ teeprom + c@ ;

\ Configuration word.

variable configword
3fff configword !

: config-mask invert configword @ and or configword ! ;

meta

: set-fosc 3 config-mask ;
0 constant fosc-lp
1 constant fosc-xt
2 constant fosc-hs
3 constant fosc-rc

: set-wdte if 4 else 0 then 4 config-mask ;
: set-/pwrte if 8 else 0 then 8 config-mask ;
: set-boden if 40 else 0 then 40 config-mask ;
: set-lvp if 80 else 0 then 80 config-mask ;
: set-cpd if 100 else 0 then 100 config-mask ;
: set-wrt if 200 else 0 then 200 config-mask ;
: set-debug if 800 else 0 then 800 config-mask ;

: set-cp 3030 config-mask ;
3030 constant no-cp
0 constant full-cp

\ ----------------------------------------------------------------------
\ Compiler logic
\ ----------------------------------------------------------------------

host

: set true swap ! ;
: unset false swap ! ;

variable deadcode?
variable code-depth
: add-depth code-depth +! ;
: inc-depth 1 add-depth ;
: dec-depth -1 add-depth ;
: check-code-area tcshere 800 >= abort" out of code space" ;
: org to tcshere check-code-area ;
: cs, deadcode? @ if drop exit then tcshere tcs! tcshere 1+ org inc-depth ;
: prevlastcs tcshere 2 - tcs@ ;
: prevprevlastcs tcshere 3 - tcs@ ;
: lastcs tcshere 1- tcs@ ;
: cs-rewind tcshere 1- org dec-depth ;
: cs-rewind2 cs-rewind cs-rewind ;
: cs-rewind3 cs-rewind2 cs-rewind ;
: cs-unwind lastcs cs-rewind ;
: cs-unwind2 cs-unwind cs-unwind ;
: cs-unwind3 cs-unwind2 cs-unwind ;
: opt? 0 code-depth @ < ;
: opt2? 1 code-depth @ < ;
: opt3? 2 code-depth @ < ;

variable tcompile
: +tcompile tcompile set ;
: -tcompile tcompile unset ;
: tcompile? tcompile @ ;

\ Incorporate immediate words from forth so that we can use them
\ while in target mode

meta

: ( postpone ( ;
: \ postpone \ ;
import: meta
import: host

\ Include common base change commands

: decimal decimal ;
: hex hex ;

host

\ The forth> word eases debugging. You can write forth> expression ('til
\ the end of line), and it will get evaluated with the regular forth
\ dictionary placed first. The meta> word does the same thing with the
\ metacompiler vocabulary and is immediate as it is used in other
\ definitions. The target> word does the same thing with the target
\ vocabulary first.

: evaluate-eol source >in @ /string evaluate postpone \ ;
: forth> also forth evaluate-eol previous ;
: meta> also metacompiler evaluate-eol previous ; immediate
: target> also picforth evaluate-eol previous ; immediate

meta

import: forth>

\ ----------------------------------------------------------------------
\ Prefix assembler, with postfix mode
\ ----------------------------------------------------------------------

picasm

variable prefix?
: prefix prefix? set ;
: postfix prefix? unset ;
: >prefix prefix? @ if evaluate-eol then ;

: ] +tcompile also picforth ;
: [ -tcompile previous ;

\ Parameterless opcodes
: clrw   103 cs, ;
: clrwdt 064 cs, ;
: nop    000 cs, ;
: retfie 009 cs, ;
: return 008 cs, ;
: sleep  063 cs, ;

\ Byte oriented file register operations
0 value f?
: ,f [ 1 7 lshift ] literal to f? ;
: ,w 0 to f? ;
: bofro: create 8 lshift , does> @ >prefix swap 7f and or f? or cs, ;
7 bofro: addwf
5 bofro: andwf
9 bofro: comf
3 bofro: decf
b bofro: decfsz
a bofro: incf
f bofro: incfsz
4 bofro: iorwf
8 bofro: movf
d bofro: rlf
c bofro: rrf
2 bofro: subwf
e bofro: swapf
6 bofro: xorwf
: clrf  >prefix 7f and 180 or cs, ;
: movwf >prefix 7f and 080 or cs, ;

\ Bit oriented file register operations
: bfro: create 4 or A lshift ,
        does> @ >prefix swap 7 lshift or swap 7f and or cs, ;
0 bfro: bcf
1 bfro: bsf
2 bfro: btfsc
3 bfro: btfss

\ Literal operations
: lo: create 8 lshift , does> @ >prefix swap ff and or cs, ;
3e lo: addlw
39 lo: andlw
38 lo: iorlw
30 lo: movlw
34 lo: retlw
3c lo: sublw
3a lo: xorlw

\ Control operations
: co: create 8 lshift , does> @ >prefix swap 7ff and or cs, ;
20 co: call
28 co: goto

\ Literals and constants

: opf? opt2? if lastcs 0800 = prevlastcs 30ff and 80 = and exit then false ;
: pushw? opt2? if lastcs 0080 = prevlastcs 0384 = and exit then false ;
: pop
    pushw? if cs-rewind2 exit then
    opf? if cs-rewind cs-unwind 80 xor cs, then
    4 ,f incf ;
: w=tos? opt? if lastcs 0800 = lastcs 0080 = or exit then false ;

: loadw w=tos? if exit then 0 ,w movf ;
: storew w=tos? if exit then 0 movwf ;
: push 4 ,f decf ;
: pushw push 0 movwf ;
: popw loadw pop ;

meta

\ Import those in the meta vocabulary

import: push     import: pop       import: loadw
import: storew   import: pushw     import: popw

: >w popw ;
: w> pushw ;

host

: literal? ( -- n ) tcshere tcslit@ ;
: literal! ( n -- ) tcshere tcslit! ;
: (literal)
    dup ff and if
	dup literal! ff and movlw meta> pushw
    else
	literal! meta> push
	0 clrf
    then ;

meta

: literal (literal) ;

host

: (t-constant)
    @ tcompile? if
	(literal)
    else
	state @ if
	    postpone literal
	then
    then ;

meta

: constant
    here >r , : r> postpone literal postpone (t-constant) postpone ;
    immediate ;

: [ -tcompile also forth ;
: ] +tcompile previous ;

target

\ Common constants

-1 constant true
0 constant false

  00 constant  indf
  01 constant  tmr0
  02 constant  pcl
  03 constant  status
  04 constant  fsr
  05 constant  porta
  06 constant  portb
  07 constant  portc
  08 constant  portd
  09 constant  porte
  0a constant  pclath
  0b constant  intcon
  0c constant  pir1
  0d constant  pir2
  0e constant  tmr1l
  0f constant  tmr1h
  10 constant  t1con
  11 constant  tmr2
  12 constant  t2con
  13 constant  sspbuf
  14 constant  sspcon
  15 constant  ccpr1l
  16 constant  ccpr1h
  17 constant  ccp1con
  18 constant  rcsta
  19 constant  txreg
  1a constant  rcreg
  1b constant  ccpr2l
  1c constant  ccpr2h
  1d constant  ccp2con
  1e constant  adresh
  1f constant  adcon0

  81 constant  option_reg
  85 constant  trisa
  86 constant  trisb
  87 constant  trisc
  88 constant  trisd
  89 constant  trise
  8c constant  pie1
  8d constant  pie2
  8e constant  pcon
  91 constant  sspcon2
  92 constant  pr2
  93 constant  sspadd
  94 constant  sspstat
  98 constant  txsta
  99 constant  spbrg
  9e constant  adresl
  9f constant  adcon1

 10c constant  eedata
 10d constant  eeadr
 10e constant  eedath
 10f constant  eeadrh

 18c constant  eecon1
 18d constant  eecon2

( status bits )

  07 constant  irp
  06 constant  reg-rp1
  05 constant  reg-rp0
  04 constant  not_to
  03 constant  not_pd
  02 constant  z
  01 constant  dc
  00 constant  c

( intcon bits )

  07 constant  gie
  06 constant  eeie
  05 constant  t0ie
  04 constant  inte
  03 constant  rbie
  02 constant  t0if
  01 constant  intf
  00 constant  rbif

( option bits )

  07 constant  /rbpu
  06 constant  intedg
  05 constant  t0cs
  04 constant  t0se
  03 constant  psa
  02 constant  ps2
  01 constant  ps1
  00 constant  ps0

( eecon1 bits )

  07 constant  eepgd
  03 constant  wrerr
  02 constant  wren
  01 constant  wr
  00 constant  rd

( txsta bits )

  07 constant  csrc
  06 constant  tx9
  05 constant  txen
  04 constant  sync
  02 constant  brgh
  01 constant  trmt
  00 constant  tx9d

( rcsta bits )

  07 constant  spen
  06 constant  rx9
  05 constant  sren
  04 constant  cren
  02 constant  ferr
  01 constant  oerr
  00 constant  rx9d

( pir1 bits )

  07 constant  pspif
  06 constant  adif
  05 constant  rcif
  04 constant  txif
  03 constant  sspif
  02 constant  ccp1if
  01 constant  tmr2if
  00 constant  tmr1if

  ( t1con bits )

  05 constant  t1ckps1
  04 constant  t1ckps0
  03 constant  t1oscen
  02 constant  /t1sync
  01 constant  tmr1cs
  00 constant  tmr1on

meta

\ Data and variables

: create
    create there , immediate
does>
    @ tcompile? if
	(literal)
    else
	state @ if
	    postpone literal
	then
    then
;

: ,
    there t!
    there 1+ dataorg
;

: c,          \ 8 bits system, "c," and "," are equivalent
    meta> ,
;

: eecreate
    create teehere , immediate
does>
    @ tcompile? if
	(literal)
    else
	state @ if
	    postpone literal
	then
    then
;

: eeallot ( n -- ) teehere + to teehere ;

: ee, ( b -- )
    teehere tee!
    teehere 1+ to teehere
;

: align ;     \ 8 bits system, no alignment is needed

: variable
    meta> create 0 ,
;
  
host

: no-opt 0 code-depth ! ;
: reachable deadcode? unset no-opt ;
: unreachable deadcode? set ;

\ In host memory, a word contains:
\    - address in target memory (1 cell)
\    - length in target memory (1 cell)
\    - address of previous word
\    - pointer to the name (counted string)
\    - depth of the code
\    - is W-TOS convention used (1 cell)
\ -1 means that length has not been computed

\ Store string as a counted string in here
: s, ( caddr n -- addr ) here >r dup c, 0 ?do dup c@ c, char+ loop drop r> ;

\ Store name (without consuming it) and return address of a counted string
: store-name ( "name" -- addr ) >in @ name s, swap >in ! ;

\ Accessors
: t-depth ( addr -- ) 4 cells + @ ;
: t-name ( addr -- ) 3 cells + @ count ;
: t-previous ( addr -- addr' ) 2 cells + @ ;
: t-length ( addr -- n ) 1 cells + @ ;

\ Control of the word being currently defined

variable last-word
: last-addr ( -- ) last-word @ ;
: start-word ( -- ) here last-word ! ;
: compute-length ( -- ) tcshere last-addr @ - last-addr cell+ ! ;
: >host ( addr -- addr'|0 )
    last-addr
    begin
	dup while
	2dup @ = if nip exit then
	t-previous
    repeat
    nip ;
: compute-depth ( -- )
    0 last-addr @ last-addr t-length over + swap do
	i tcs@ 3000 and 2000 = if
	    i tcs@ 7ff and >host ?dup if
		t-depth
		i tcs@ 800 and 0= if 1+ then
		max
	    then
	then
    loop
    last-addr 4 cells + !
    ;
: end-word ( -- ) compute-length compute-depth ;

: (t-act) 
    tcompile? if
	dup 5 cells + @ if popw then
	@ call
    else
	@ state @ if
	    postpone literal
	then
    then
;

: t-header ( w? -- )
    store-name create last-addr start-word tcshere , -1 , , , -1 , ,
    reachable
does> (t-act)
;

meta

: ]asm postfix -tcompile also picassembler ;
: asm[ +tcompile previous reachable ;

: code false t-header ]asm ;

: nip meta> popw storew ;

: recurse last-addr @ call ;

host

: call? opt? if lastcs 3800 and 2000 = exit then false ;
: call-address cs-unwind 7ff and ;

: const?-0 opt2? if
    lastcs 0180 = prevlastcs 0384 = and exit
  else false then ;

: const?-non-0 opt3? if
    lastcs 0080 = prevlastcs 0384 = and prevprevlastcs 3800 and 3000 = and
  else false then ;

: const? const?-0 const?-non-0 or ;
: kill-const const?-0 if cs-rewind2 literal?
  else cs-rewind3 literal? then ;

meta

: +
    const? if
	kill-const const? if
	    kill-const + (literal)
	else
	    pushw? if
		popw addlw pushw
	    else
		movlw
		indf ,f addwf
	    then
	then
    else
	popw
	indf ,f addwf
    then
;

: -
    const? if
	kill-const const? if
	    kill-const swap - (literal)
	else
	    pushw? if
		popw invert addlw pushw
	    else
		invert movlw
		indf ,f addwf
	    then
	then
    else
	popw indf ,f subwf
    then
;

: allot 0 ?do meta> 0 c,
  loop ;

target

\ Create a virtually infinite (16 slots) data stack
create picforth-stack 10 allot

\ Temporary variables

variable tmp1
variable tmp2
variable tmp3

host

\ Words to manipulate temporary variables

: w>tmp1 [ tmp1 ] literal movwf ;
: tmp1>w [ tmp1 ] literal ,w movf ;
: w>tmp2 [ tmp2 ] literal movwf ;
: tmp2>w [ tmp2 ] literal ,w movf ;
: w>tmp3 [ tmp3 ] literal movwf ;
: tmp3>w [ tmp3 ] literal ,w movf ;

\ Words to manipulate fsr register

: w>fsr [ fsr ] literal movwf ;
: fsr>w [ fsr ] literal ,w movf ;

meta

import: w>tmp1   import: tmp1>w
import: w>tmp2   import: tmp2>w
import: w>tmp3   import: tmp3>w
import: w>fsr    import: fsr>w

host

\ Modules loaded on-demand. The current solution is not very efficient
\ but it works.

variable (!)-loaded
: (!)
    (!)-loaded @ dup if execute exit then
    1 abort" Please include libstore.fs in your application"
;

variable (@)-loaded
: (@)
    (@)-loaded @ dup if execute exit then
    1 abort" Please include libfetch.fs in your application"
;

variable current-bank

: rp0bsf ( -- )
    opt? if lastcs 1283 = if cs-rewind exit then then
    opt2? if lastcs ff00 and 3000 = prevlastcs 1283 = and if
	cs-unwind cs-rewind cs, exit
    then then
    status reg-rp0 bsf
;

: rp1bsf ( -- )
    opt? if lastcs 1303 = if cs-rewind exit then then
    opt2? if lastcs ff00 and 3000 = prevlastcs 1303 = and if
	cs-unwind cs-rewind cs, exit
    then then
    status reg-rp1 bsf
;

: adjust-bank ( a -- a' )
    case dup 180 and dup current-bank !
	80  of rp0bsf endof
	100 of rp1bsf endof
	180 of rp0bsf rp1bsf endof
    endcase
    7f and
;

: rp0bcf ( -- ) status reg-rp0 bcf ;
: rp1bcf ( -- ) status reg-rp1 bcf ;

: restore-bank ( -- )
    case current-bank @
	80  of rp0bcf endof
	100 of rp1bcf endof
	180 of rp1bcf rp0bcf endof
    endcase
    0 current-bank !
;

: const-! ( addr -- )
    const? if
	kill-const dup 0= if
	    drop adjust-bank clrf restore-bank
	else
	    movlw adjust-bank movwf restore-bank
	then
    else
	meta> popw adjust-bank movwf restore-bank
    then
;

meta

import: adjust-bank   import: restore-bank

: !
    const? if
	kill-const const-!
    else
	(!)
    then
;

: @
    const? if
	kill-const ,w adjust-bank movf restore-bank meta> pushw
    else
	(@)
    then
;

: c@ meta> @
;

: c! meta> !
;

: swap
    popw
    indf ,w xorwf
    indf ,f xorwf
    indf ,w xorwf
    pushw
;

: over popw w>tmp1 loadw w>tmp2 tmp1>w pushw tmp2>w pushw ;

: tuck popw w>tmp1 loadw w>tmp2 tmp1>w storew tmp2>w pushw tmp1>w pushw ;

: 2dup popw w>tmp2 loadw w>tmp1 pushw tmp2>w pushw tmp1>w pushw tmp2>w pushw ;

: +!
    const? if                    \ Address is constant
	kill-const
	const? if                \ Number is constant
	    kill-const
	    dup 1 = if           \ Number is one
		drop adjust-bank ,f incf restore-bank
	    else dup ff = if     \ Number is -1
		drop adjust-bank ,f decf restore-bank
	    else
		movlw adjust-bank ,f addwf restore-bank
	    then then
	else
	    popw adjust-bank addwf restore-bank
	then
    else
	meta> tuck @ + swap !
    then
;

: negate
    const? if kill-const negate (literal) exit then
    pushw? if popw 0 sublw pushw exit then
    indf ,f comf
    indf ,f incf
;

: invert
    const? if kill-const invert (literal) exit then
    pushw? if popw ff xorlw pushw exit then
    indf ,f comf
;

host

: pop-value-w const? if kill-const movlw else popw then ;

meta

: 1+ pushw? if 1 (literal) meta> +
     else indf ,f incf then ;
: 1- pushw? if 1 (literal) meta> -
     else indf ,f incf then ;

\ Flow control words

forth

: resolve ( faddr -- ) tcshere swap org reachable dup goto org reachable ;

meta

: ahead ( -- faddr ) tcshere 0 goto unreachable ;

: complastcs lastcs ff00 and = ;

: zbit?
    opt? 0= if false exit then
    3900 complastcs if true exit then   \ andlw
    0500 complastcs if true exit then   \ andwf
    0800 complastcs if true exit then   \ movlw
    3e00 complastcs if true exit then   \ addlw
    0700 complastcs if true exit then   \ addwf
    lastcs 103 = if true exit then      \ clrw
    0900 complastcs if true exit then   \ comf
    0300 complastcs if true exit then   \ decf
    0a00 complastcs if true exit then   \ incf
    3800 complastcs if true exit then   \ iorlw
    0400 complastcs if true exit then   \ iorwf
    3c00 complastcs if true exit then   \ sublw
    0200 complastcs if true exit then   \ subwf
    3a00 complastcs if true exit then   \ xorlw
    0600 complastcs if true exit then   \ xorwf
    0800 complastcs if true exit then   \ movf
    false ;
 
: test-z-bit ( -- )
    zbit? 0= if 0 iorlw then
;

: test-structure ( -- faddr )
    meta> ahead
    reachable
;

: log2 ( n -- n' )
    7 for 1 lshift dup 100 = if drop i unloop exit then next drop -1 ;

: bit-test? ( -- f )
    opt2? if lastcs ff00 and 3900 =
	     lastcs ff and log2 -1 <> and
	     prevlastcs ff00 and 0800 = and
	 else
	     false
	 then
;

: rewrite-bit-test ( -- )
    cs-unwind2 ff and swap ff and log2 ;

: if ( -- faddr )
    popw
    bit-test? if rewrite-bit-test btfss else status z btfsc then
    test-structure
;

: /if ( -- faddr )
    popw
    test-z-bit
    bit-test? if rewrite-bit-test btfsc else status z btfss then
    test-structure
;

\ At resolve time, check whether we have something like:
\   btfsc ... or btfss ...
\   goto  0 \ unresolved yet
\   any instruction
\ If this is the case, this will be changed into:
\   btfss ... or btfsc ...
\   any instruction
\ and no further action will be taken.

: short-if? ( faddr -- faddr f )
    dup 2 + tcshere =
    prevlastcs 2800 = and
    prevprevlastcs 1800 and 1800 = and
;
    
: then ( faddr -- )
    short-if? if
	drop reachable cs-unwind cs-rewind cs-unwind 0400 xor cs, cs,
	reachable
    else
	resolve
    then ;
    
: else ( faddr1 -- faddr2 )
    meta> ahead
    swap
    meta> then
;

: backref ( -- baddr ) tcshere no-opt ;

: begin ( -- 0 baddr ) 0 backref ;

: again ( 0 baddr -- ) goto drop unreachable ;

: while ( 0 i*x baddr -- 0 i*x faddr baddr )
    meta> if
    swap
;

: /while ( 0 i*x baddr -- 0 i*x faddr baddr )
    meta> /if
    swap
;

: repeat ( 0 i*x baddr -- )
    0 swap meta> again
    begin dup while meta> then
    repeat drop
;

: until ( 0 baddr -- )
    meta> /while repeat
;

: /until ( 0 baddr -- )
    meta> while repeat
;

\ Some of thoses are too lengthy to be used inline and should be put in an
\ auto-loading library. Maybe with an "import" keyword or something like that.

: 0=
    clrw
    indf ,f iorwf
    status z btfsc
    ff addlw
    indf movwf
;

: 0<>
    clrw
    indf ,f iorwf
    status z btfss
    ff addlw
    indf movwf
;

: =
    meta> - 0=
;

: <>
    meta> - 0<>
;

: carry-set?
    clrw     \ Does not affect carry bit
    status c btfsc
    ff addlw
    indf movwf
;

: carry-clear?
    clrw     \ Does not affect carry bit
    status c btfss
    ff addlw
    indf movwf
;

: <=
    meta> -
    carry-clear?
;

: >
    meta> -
    carry-set?
;

\ The next two are awful but are included for completeness

: <
    meta> swap >
;

: >=
    meta> swap <=
;

meta

: get-const ( -- const )
    const? 0= if -1 abort" can only be used with a constant" then
    kill-const ;

: v-for ( -- baddr ) ( runtime: n -- )
    meta> ! begin
;

: v-next ( baddr -- ) ( runtime: -- )
    get-const ,f decfsz meta> again
    reachable
;

: table ( -- )
    popw pcl ,f addwf ;

: tc: ( -- )
    reachable ' >body @ goto ; unreachable

: dup loadw pushw ;
: drop pop ;
: 2drop pop pop ;

: and
    const? if
	kill-const const? if
	    kill-const and (literal)
	else
	    pushw? if cs-rewind2 andlw pushw
	    else movlw indf ,f andwf then
	then
    else
	popw indf ,f andwf
    then
;

: and!
    const? if
	kill-const const? if
	    kill-const dup invert ff and log2 dup -1 <> if
		nip >r adjust-bank r> bcf restore-bank
	    else
		drop movlw ,f adjust-bank andwf restore-bank
	    then
	else
	    pop-value-w ,f adjust-bank andwf restore-bank
	then
    else
	meta> tuck @ and swap !
    then
;

: /and!
    const? if
	kill-const meta> invert literal and!
    else
	meta> swap invert swap and!
    then
;

: xor
    const? if
	kill-const const? if
	    kill-const xor (literal)
	else
	    pushw? if cs-rewind2 xorlw pushw
	    else movlw indf ,f xorwf then
	then
    else
	popw indf ,f xorwf
    then
;

: xor!
    const? if
	kill-const pop-value-w ,f adjust-bank xorwf restore-bank
    else
	meta> tuck @ xor swap !
    then
;

: or
    const? if
	kill-const const? if
	    kill-const or (literal)
	else
	    pushw? if
		cs-rewind2 iorlw pushw
	    else
		dup log2 dup -1 <> if
		    indf swap bsf drop
		else
		    drop
		    movlw indf ,f iorwf
		then
	    then
	then
    else
	popw indf ,f iorwf
    then
;

: or!
    const? if
	kill-const const? if
	    kill-const dup log2 dup -1 <> if
		nip >r adjust-bank r> bsf restore-bank
	    else
		drop movlw ,f adjust-bank iorwf restore-bank
	    then
	else
	    pop-value-w ,f adjust-bank iorwf restore-bank
	then
    else
	meta> tuck @ or swap !
    then
;

: bit-set
    get-const get-const swap adjust-bank swap bsf restore-bank ;
: bit-clr
    get-const get-const swap adjust-bank swap bcf restore-bank ;
: bit-toggle
    get-const 1 get-const lshift movlw ,f adjust-bank xorwf restore-bank ;
: bit-set?
    get-const 1 get-const lshift swap meta> (literal) @ (literal) and
    ;

: >input get-const 80 or (literal) bit-set ;
: >output get-const 80 or (literal) bit-clr ;
    
\ Aliases, more readable for port manipulation
    
: high
    meta> bit-set
;
: low
    meta> bit-clr
;
: high?
    meta> bit-set?
;
: toggle
    meta> bit-toggle
;

: pin-a ( n -- ) create , does> @ (literal) porta (literal) ;
: pin-b ( n -- ) create , does> @ (literal) portb (literal) ;
: pin-c ( n -- ) create , does> @ (literal) portc (literal) ;
: pin-d ( n -- ) create , does> @ (literal) portd (literal) ;
: pin-e ( n -- ) create , does> @ (literal) porte (literal) ;
    
: rlf-tos indf ,f rlf ;
: rrf-tos indf ,f rrf ;

: clrwdt clrwdt ;

\ Definitions

: : false t-header +tcompile ;
: :: true t-header +tcompile pushw ;
: (return)
    call? if call-address goto exit then
    return ;
: (exit) (return) unreachable ;
: exit (exit) ;
: ; (exit) -tcompile end-word ;

: label: ( "name" -- ) create tcshere , does> (t-act) ;

: ' ( "name" -- xt )
    ' >body @ ;

: init-picforth
    s" : (init-picforth)" evaluate
    [ picforth-stack 10 + ] literal movlw fsr movwf
    end-word
;

\ Set a goto to the current address at addr
: set-vector ( addr -- ) tcshere swap org reachable dup goto org reachable ;

\ Address of main program
: main ( -- ) 0 set-vector init-picforth ;

\ Reset the device -- caution must be taken: not everything will return
\ in default state, only the program counter and the stacks (well, only
\ the data stack, but since the return stack is a circular buffer, it has
\ the same effect for any correct program).
: reset ( -- ) 0 goto ;

\ Dummy interrupt handling routines (overloaded when picisr.fs is loaded)
: enable-interrupts ( -- ) ;
: disable-interrupts ( -- ) ;

picasm

: end-code previous end-word ;

host

\ ----------------------------------------------------------------------
\ Outer interpreter
\ ----------------------------------------------------------------------

\ This is needed as numbers cannot have special treatments on a
\ regular Forth without building another outer interpreter.

: throw-unknown ( caddr n -- )
    ." Unknown word: " type cr abort
;

: do-number ( n -- |n )
    state @ if postpone literal exit then
    tcompile? if [ meta ] (literal) [ host ] then
;

: >number ( caddr n -- )
    2dup 2>r snumber? dup 0= if drop 2r> throw-unknown then
    2r> 2drop
    1 = if do-number then
    do-number
;

: parser ( caddr n -- )
    2dup sfind
    dup 0= if drop >number exit then
    2swap 2drop 1 = state @ 0= or if execute else compile, then
;

: interpret ( -- )
    begin ?stack name dup while parser repeat
    2drop ;

: pic'quit ( -- ) begin .status cr (query) interpret prompt again ;

\ The following comes from gforth

: picquit ( ?? -- ?? )
    \ Empty the return stack, make the user input device
    \ the input source, enter interpret state and start
    \ the text interpreter.
    rp0 @ rp! handler off clear-tibstack >tib @ >r
    \ Enter target mode
    target
    begin
        postpone [
        ['] pic'quit catch dup
    while
        <# \ reset hold area, or we may get another error
        DoError r@ >tib ! r@ tibstack !
    repeat
    drop r> >tib ! ;

: read-loop ( i*x -- j*x )
    begin refill while interpret repeat ;

: include-file1 ( i*x wfileid -- j*x ior1 ior2 )
    \ Interpret (process using the text interpreter) the contents of
    \ the file @var{wfileid}.
    push-file  loadfile !
    0 loadline ! blk off  ['] read-loop catch
    loadfile @ close-file swap 2dup or
    pop-file  drop ;

: include-file2 ( i*x wfileid -- j*x )
    \ like include-file, but does not update loadfile#
    include-file1 throw throw ;

: included1 ( i*x file-id c-addr u -- j*x ) \ gforth
    \ Include the file file-id with the name given by @var{c-addr u}.
    loadfilename# @ >r
    save-mem add-included-file ( file-id )
    included-files 2@ nip 1- loadfilename# !
    ['] include-file2 catch
    r> loadfilename# !
    throw ;
    
: included ( i*x c-addr u -- j*x ) \ file
    \ include-file the file whose name is given by the string
    \ c-addr u.
    open-fpath-file throw included1 ;

: include  ( ... "file" -- ... ) \ gforth
    name included ;

\ end of gforth originated code

meta

\ Add the capability to include a file

import: include

\ ----------------------------------------------------------------------
\ Dump of hexadecimal data
\ ----------------------------------------------------------------------

host

-1 value fd
: fd? fd -1 <> ;
: emit-fd fd? if fd emit-file throw else emit then ;
: type-fd fd? if fd write-file throw else type then ;
create crlf a c, d c,
: cr-fd fd? if a emit-fd else cr then ;

variable cks
: +cks cks +! ;
: hex8. dup +cks s>d <# # # #> type-fd ;
: hex16. dup 8 rshift hex8. ff and hex8. ;
: le-hex16. dup ff and hex8. 8 rshift hex8. ;
: mem-dump tcs@ le-hex16. ;
: start-line [char] : emit-fd 0 cks ! ;
: end-line cks @ ff and negate hex8. cr-fd ;
: line-dump ( addr -- )
    start-line 10 hex8. dup 2* hex16. 0 hex8.
    7 for dup mem-dump 1+ next drop end-line ;
: code-dump
    0 begin dup tcshere < while dup line-dump 8 + repeat drop ;
: ee-dump tee@ le-hex16. ;
: eeline-dump ( addr -- )
    start-line 10 hex8. dup 2100 + 2* hex16. 0 hex8.
    7 for dup ee-dump 1+ next drop end-line ;
: eeprom-dump
    0 begin dup teehere < while dup eeline-dump 8 + repeat drop ;
: end-dump s" :00000001FF" type-fd cr-fd ;
: config-dump
    start-line 2 hex8. 400E hex16. 0 hex8. configword @ le-hex16. end-line ;
: dump hex code-dump config-dump eeprom-dump end-dump ;

: dump-file ( caddr a -- )
    w/o create-file throw to fd dump fd close-file throw ;

: file-dump ( "name" -- ) name dump-file ;

meta

import: file-dump
import: bye

\ ----------------------------------------------------------------------
\ PIC disassembler
\ ----------------------------------------------------------------------

forth

\ Printable output

: .tab 9 emit ;
: (.op) ( caddr n -- ) .tab type .tab ;
: .op" postpone s" postpone (.op) ; immediate \ "
: .lit ." 0x" ff and s>d <# # # #> type ;
: .comment .tab ." ; " ;
: .file ." 0x" 7f and s>d <# # # #> type ;
: .addr ." 0x" s>d <# # # # #> type ;
: .4const s>d <# # # # # #> type ;
: .4hexconst ." 0x" .4const ;
: .bconst [char] 0 + emit ;
: .filewf dup .file ." ," 80 and if ." f" else ." w" then ;

\ Unknown opcode

: dis-unknown ( w -- ) .op" .dw" .4hexconst ;

\ Byte-oriented file register operations

: dis-0 ( w -- )
    dup 80 and if .op" movwf" .file exit then
    dup 1f and 0= if .op" nop" drop exit then
    dup 64 = if drop .op" clrwdt" exit then
    dup 9 = if drop .op" retfie" exit then
    dup 8 = if drop .op" return" exit then
    dup 63 = if drop .op" sleep" exit then
    dis-unknown
;

: dis-1 ( w -- )
    dup 80 and if .op" clrf" .file exit then
    .op" clrw" drop
;

: dis-fileop ( w -- )
    case dup f00 and
	000 of dis-0 endof
	100 of dis-1 endof
	200 of .op" subwf" .filewf endof
	300 of .op" decf" .filewf endof
	400 of .op" iorwf" .filewf endof
	500 of .op" andwf" .filewf endof
	600 of .op" xorwf" .filewf endof
	700 of .op" addwf" .filewf endof
	800 of .op" movf" .filewf endof
	900 of .op" comf" .filewf endof
	a00 of .op" incf" .filewf endof
	b00 of .op" decfsz" .filewf endof
	c00 of .op" rrf" .filewf endof
	d00 of .op" rlf" .filewf endof
	e00 of .op" swapf" .filewf endof
	f00 of .op" incfsz" .filewf endof
    endcase
;

\ Bit-oriented file register operations

: dis-bitop ( w -- )
    case dup c00 and
	000 of .op" bcf" endof
	400 of .op" bsf" endof
	800 of .op" btfsc" endof
	c00 of .op" btfss" endof
    endcase
    dup .file ." ," 380 and 7 rshift .bconst
;

\ Control operations

: .addr-name ( addr -- )
    last-addr
    begin
	dup while
	2dup @ >= if
	    dup t-name type 2dup @ - ?dup if
		."  + " .addr 2drop
	    else
		t-depth ?dup if
		    ."  (rs depth: " .bconst ." )"
		then
		drop
	    then
	    exit
	then
	t-previous
    repeat
    2drop ;

: dis-control ( w -- )
    dup 800 and if .op" goto" else .op" call" then
    7ff and dup .addr .comment .addr-name
;

\ Literal operations

: dis-lit ( w -- )
    dup e00 and e00 = if .op" addlw" .lit exit then
    dup f00 and 900 = if .op" andlw" .lit exit then
    dup f00 and 800 = if .op" iorlw" .lit exit then
    dup c00 and 0= if .op" movlw" .lit exit then
    dup c00 and 400 = if .op" retlw" .lit exit then
    dup e00 and c00 = if .op" sublw" .lit exit then
    dup f00 and a00 = if .op" xorlw" .lit exit then
    3000 or dis-unknown
;

\ Disassemble word

: dis-v ( w -- )
    case dup 3000 and
	0000 of dis-fileop endof
	1000 of dis-bitop endof
	2000 of dis-control endof
	3000 of dis-lit endof
    endcase
    cr
;

: dis-a ( addr -- )
    dup >host ?dup if
	.comment ." name: " dup t-name type cr
	.comment ." max return-stack depth: " t-depth .bconst cr
    then
    dup .4hexconst .tab tcs@ dup .4const dis-v ;

meta

: see ( "name" -- ) cr ' >body 2@ over 0 do dup dis-a 1+ loop drop ;

: words ( -- )
    cr last-addr
    begin dup while dup t-name type space t-previous repeat
    cr ;

: map ( -- )
    last-addr
    begin
	dup while
	dup @ .addr .tab dup t-depth .bconst .tab dup t-name type .tab 
	t-previous cr
    repeat
;

: dis ( -- ) tcshere 0 ?do i dis-a loop ;

\ Switch into picforth parser mode from now on

target

